"""The `call_factory` method for generating provider specific call decorators."""

from enum import Enum
from functools import partial
from typing import (
    Annotated,
    AsyncIterable,
    Callable,
    Iterable,
    Literal,
    NoReturn,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ._create import create_factory
from ._extract import extract_factory
from ._stream import BaseStream, stream_factory
from ._structured_stream import structured_stream_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    HandleStream,
    HandleStreamAsync,
    LLMFunctionDecorator,
    SetupCall,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_ResponseModelT = TypeVar(
    "_ResponseModelT", bound=BaseModel | BaseType | Enum | Annotated
)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")


def call_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TDynamicConfig: type[_BaseDynamicConfigT],
    TToolType: type[_BaseToolT],
    TStream: type[_BaseStreamT],
    TCallParams: type[_BaseCallParamsT],
    default_call_params: _BaseCallParamsT,
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT | _BaseCallResponseChunkT],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
):
    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig, TCallResponse, TCallResponse
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig, _ParsedOutputT, _ParsedOutputT
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[TDynamicConfig, TStream, TStream]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig, _ResponseModelT, _ResponseModelT
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig, _ParsedOutputT, _ParsedOutputT
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True],
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig, Iterable[_ResponseModelT], AsyncIterable[_ResponseModelT]
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True],
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[TCallResponse], _ParsedOutputT]
        | Callable[[TCallResponseChunk], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> NoReturn: ...  # pragma: no cover

    def base_call(
        model: str,
        *,
        stream: bool = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT]
        | Callable[[TCallResponseChunk], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        call_params: TCallParams = default_call_params,
    ) -> LLMFunctionDecorator[
        TDynamicConfig,
        TCallResponse
        | _ParsedOutputT
        | TStream
        | _ResponseModelT
        | Iterable[_ResponseModelT],
        TCallResponse
        | _ParsedOutputT
        | TStream
        | _ResponseModelT
        | AsyncIterable[_ResponseModelT],
    ]:
        if stream and output_parser:
            raise ValueError("Cannot use `output_parser` with `stream=True`.")

        if response_model:
            if stream:
                return partial(
                    structured_stream_factory(
                        TCallResponse=TCallResponse,
                        TCallResponseChunk=TCallResponseChunk,
                        TStream=TStream,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # type: ignore
            else:
                return partial(
                    extract_factory(
                        TCallResponse=TCallResponse,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                    ),
                    model=model,
                    response_model=response_model,
                    output_parser=output_parser,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )  # type: ignore
        if stream:
            return partial(
                stream_factory(
                    TCallResponse=TCallResponse,
                    TStream=TStream,
                    setup_call=setup_call,
                    handle_stream=handle_stream,
                    handle_stream_async=handle_stream_async,
                ),
                model=model,
                tools=tools,
                json_mode=json_mode,
                client=client,
                call_params=call_params,
            )  # type: ignore
        return partial(
            create_factory(TCallResponse=TCallResponse, setup_call=setup_call),
            model=model,
            tools=tools,
            output_parser=output_parser,
            json_mode=json_mode,
            client=client,
            call_params=call_params,
        )  # type: ignore

    return base_call



"""This module contains the `get_template_values` function."""

from typing import Any


def get_template_values(
    template_variables: list[tuple[str, str | None]], attrs: dict[str, Any]
) -> dict[str, Any]:
    """Returns the values of the given `template_variables` from the provided `attrs`.

    Args:
        template_variables: The variables to extract from the `attrs`.
        attrs: The attributes to extract the variables from.

    Returns:
        The values of the template variables.
    """
    values = {}
    if "self" in attrs:
        values["self"] = attrs.get("self")
    for var, format_spec in template_variables:
        if var.startswith("self"):
            values["self"] = attrs.get("self")
        elif "." in var:
            var = var.split(".")[0]
            values[var] = attrs.get(var)
            continue
        elif format_spec in ["list", "lists"]:
            value = attrs[var]
            if format_spec == "list":
                if not isinstance(value, list):
                    raise ValueError(
                        f"Template variable '{var}' must be a list when using the "
                        "'list' format spec."
                    )
                values[var] = "\n".join([str(item) for item in attrs[var]])
            else:
                if not isinstance(value, list) or (
                    value and not all(isinstance(item, list) for item in value)
                ):
                    raise ValueError(
                        f"Template variable '{var}' must be a list of lists when using "
                        "the 'lists' format spec."
                    )
                values[var] = "\n\n".join(
                    ["\n".join([str(subitem) for subitem in item]) for item in value]
                )
        else:
            values[var] = str(attrs[var]) if attrs[var] is not None else ""
    return values



"""Calculate the cost of a Gemini API call."""


def calculate_cost(
    input_tokens: int | float | None, output_tokens: int | float | None, model: str
) -> float | None:
    """Calculate the cost of a Gemini API call."""
    return None



"""The `Metadata` typed dictionary for including general metadata."""

from typing_extensions import NotRequired, TypedDict


class Metadata(TypedDict, total=False):
    """The `Metadata` typed dictionary for including general metadata."""

    tags: NotRequired[set[str]]



"""Vectorstores for the RAG module."""

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, Optional, TypeVar, Union

from pydantic import BaseModel

from .chunkers import BaseChunker, TextChunker
from .config import BaseConfig
from .document import Document
from .embedders import BaseEmbedder
from .query_results import BaseQueryResults
from .vectorstore_params import BaseVectorStoreParams

BaseQueryResultsT = TypeVar("BaseQueryResultsT", bound=BaseQueryResults)


class BaseVectorStore(BaseModel, Generic[BaseQueryResultsT], ABC):
    """The base class abstract interface for interacting with vectorstores."""

    api_key: ClassVar[Optional[str]] = None
    index_name: ClassVar[Optional[str]] = None
    chunker: ClassVar[BaseChunker] = TextChunker(chunk_size=1000, chunk_overlap=200)
    embedder: ClassVar[BaseEmbedder]
    vectorstore_params: ClassVar[BaseVectorStoreParams] = BaseVectorStoreParams()
    configuration: ClassVar[BaseConfig] = BaseConfig()
    _provider: ClassVar[str] = "base"

    @abstractmethod
    def retrieve(self, text: str, **kwargs: Any) -> BaseQueryResultsT:
        """Queries the vectorstore for closest match"""
        ...  # pragma: no cover

    @abstractmethod
    def add(self, text: Union[str, list[Document]], **kwargs: Any) -> None:
        """Takes unstructured data and upserts into vectorstore"""
        ...  # pragma: no cover



"""The `GroqStream` class for convenience around streaming LLM calls."""

from groq.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from groq.types.chat.chat_completion import Choice
from groq.types.chat.chat_completion_message import ChatCompletionMessage
from groq.types.completion_usage import CompletionUsage

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import GroqCallParams
from .call_response import GroqCallResponse
from .call_response_chunk import GroqCallResponseChunk
from .dynamic_config import GroqDynamicConfig
from .tool import GroqTool

FinishReason = Choice.__annotations__["finish_reason"]


class GroqStream(
    BaseStream[
        GroqCallResponse,
        GroqCallResponseChunk,
        ChatCompletionUserMessageParam,
        ChatCompletionAssistantMessageParam,
        ChatCompletionToolMessageParam,
        ChatCompletionMessageParam,
        GroqTool,
        GroqDynamicConfig,
        GroqCallParams,
        FinishReason,
    ]
):
    _provider = "groq"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self,
        tool_calls: list[ChatCompletionMessageToolCallParam] | None = None,
        content: str | None = None,
    ) -> ChatCompletionAssistantMessageParam:
        message_param = ChatCompletionAssistantMessageParam(
            role="assistant",
            content=content,
        )
        if tool_calls:
            message_param["tool_calls"] = tool_calls
        return message_param

    def construct_call_response(self) -> GroqCallResponse:
        """Constructs the call response from a consumed GroqStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        message = {
            "role": self.message_param["role"],
            "content": self.message_param.get("content", ""),
            "tool_calls": self.message_param.get("tool_calls", []),
        }
        if not self.input_tokens and not self.output_tokens:
            usage = None
        else:
            usage = CompletionUsage(
                prompt_tokens=int(self.input_tokens or 0),
                completion_tokens=int(self.output_tokens or 0),
                total_tokens=int(self.input_tokens or 0) + int(self.output_tokens or 0),
            )
        completion = ChatCompletion(
            id=self.id if self.id else "",
            model=self.model,
            choices=[
                Choice(
                    finish_reason=self.finish_reasons[0]
                    if self.finish_reasons and self.finish_reasons[0]
                    else "stop",
                    index=0,
                    message=ChatCompletionMessage.model_validate(message),
                )
            ],
            created=0,
            object="chat.completion",
            usage=usage,
        )
        return GroqCallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



import inspect
from abc import update_abstractmethods
from typing import Any, Callable, TypeVar, cast, get_type_hints

import jiter
from docstring_parser import parse
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from ._default_tool_docstring import DEFAULT_TOOL_DOCSTRING

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def convert_function_to_base_tool(
    fn: Callable,
    base: type[BaseToolT],
    __doc__: str | None = None,
    __namespace__: str | None = None,
) -> type[BaseToolT]:
    """Constructst a `BaseToolT` type from the given function.

    This method expects all function parameters to be properly documented in identical
    order with identical variable names, as well as descriptions of each parameter.
    Errors will be raised if any of these conditions are not met.

    Args:
        fn: The function to convert.
        base: The `BaseToolT` type to which the function is converted.
        __doc__: The docstring to use for the constructed `BaseToolT` type.
        __namespace__: The namespace to use for the constructed `BaseToolT` type.

    Returns:
        The constructed `BaseToolT` type.

    Raises:
        ValueError: if the given function's parameters don't have type annotations.
        ValueError: if a given function's parameter is in the docstring args section but
            the name doesn't match the docstring's parameter name.
        ValueError: if a given function's parameter is in the docstring args section but
            doesn't have a docstring description.
    """
    docstring, examples = None, []
    func_doc = __doc__ or fn.__doc__
    if func_doc:
        docstring = parse(func_doc)
        for example in docstring.examples or []:
            if example.description:
                examples.append(jiter.from_json(example.description.encode()))

    field_definitions = {}
    hints = get_type_hints(fn)
    has_self = False
    for i, parameter in enumerate(inspect.signature(fn).parameters.values()):
        if parameter.name == "self":
            has_self = True
            continue
        if parameter.name == "cls":
            continue
        if parameter.annotation == inspect.Parameter.empty:
            raise ValueError("All parameters must have a type annotation.")

        docstring_description = None
        if docstring and i < len(docstring.params):
            docstring_param = docstring.params[i]
            if docstring_param.arg_name != parameter.name:
                raise ValueError(
                    f"Function parameter name {parameter.name} does not match docstring "
                    f"parameter name {docstring_param.arg_name}. Make sure that the "
                    "parameter names match exactly."
                )
            if not docstring_param.description:
                raise ValueError("All parameters must have a description.")
            docstring_description = docstring_param.description

        field_info = FieldInfo(annotation=hints[parameter.name])
        if parameter.default != inspect.Parameter.empty:
            field_info.default = parameter.default
        if docstring_description:  # we check falsy here because this comes from docstr
            field_info.description = docstring_description

        param_name = parameter.name
        if param_name.startswith("model_"):  # model_ is a BaseModel reserved namespace
            param_name = "aliased_" + param_name
            field_info.alias = parameter.name
            field_info.validation_alias = parameter.name
            field_info.serialization_alias = parameter.name

        field_definitions[param_name] = (
            hints[parameter.name],
            field_info,
        )

    model = create_model(
        f"{__namespace__}_{fn.__name__}" if __namespace__ else fn.__name__,
        __base__=base,
        __doc__=inspect.cleandoc(func_doc) if func_doc else DEFAULT_TOOL_DOCSTRING,
        **cast(dict[str, Any], field_definitions),
    )
    if examples:
        model.model_config["json_schema_extra"] = {"examples": examples}

    def call(self: base):
        return fn(
            **(
                ({"self": self} if has_self else {})
                | {
                    str(
                        self.model_fields[field_name].alias
                        if self.model_fields[field_name].alias
                        else field_name
                    ): getattr(self, field_name)
                    for field_name in self.model_dump(exclude={"tool_call"})
                }
            )
        )

    async def call_async(self: base):
        return await call(self)

    if inspect.iscoroutinefunction(fn):
        setattr(model, "call", call_async)
    else:
        setattr(model, "call", call)
    return update_abstractmethods(model)



from typing import Any

from pydantic import BaseModel, ConfigDict


class BaseVectorStoreParams(BaseModel):
    """The parameters with which to make a vectorstore."""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def kwargs(
        self,
    ) -> dict[str, Any]:
        """Returns all parameters for the index as a keyword arguments dictionary."""
        kwargs = {
            key: value for key, value in self.model_dump().items() if value is not None
        }
        return kwargs



"""The `GeminiTool` class for easy tool usage with Google's Gemini LLM calls."""

from __future__ import annotations

from typing import Any

from google.ai.generativelanguage import FunctionCall
from google.generativeai.types import (  # type: ignore
    FunctionDeclaration,
    Tool,
)
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class GeminiTool(BaseTool):
    """A class for defining tools for Gemini LLM calls."""

    tool_call: SkipJsonSchema[FunctionCall]

    @classmethod
    def tool_schema(cls) -> Tool:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined."""
        model_schema = cls.model_tool_schema()
        fn: dict[str, Any] = {"name": cls._name(), "description": cls._description()}
        if model_schema["properties"]:
            fn["parameters"] = model_schema  # type: ignore
        if model_schema["required"]:
            fn["parameters"]["required"] = model_schema["required"]
        if "parameters" in fn:
            if "$defs" in fn["parameters"]:
                raise ValueError(
                    "Unfortunately Google's Gemini API cannot handle nested structures "
                    "with $defs."
                )

            def handle_enum_schema(prop_schema):
                if "enum" in prop_schema:
                    prop_schema["format"] = "enum"
                return prop_schema

            fn["parameters"]["properties"] = {
                prop: {
                    key: value
                    for key, value in handle_enum_schema(prop_schema).items()
                    if key != "default"
                }
                for prop, prop_schema in fn["parameters"]["properties"].items()
            }
        return Tool(function_declarations=[FunctionDeclaration(**fn)])

    @classmethod
    def from_tool_call(cls, tool_call: FunctionCall) -> GeminiTool:
        """Constructs an `GeminiTool` instance from a `tool_call`."""
        if not tool_call.args:
            raise ValueError("Tool call doesn't have any arguments.")
        model_json: dict[str, Any] = {
            key: value for key, value in tool_call.args.items()
        }
        model_json["tool_call"] = tool_call
        return cls.model_validate(model_json)



"""The default docstring to use when tools don't have a docstring."""

DEFAULT_TOOL_DOCSTRING = """\
Correctly formatted and typed parameters extracted from the completion. \
Must include required parameters and may exclude optional parameters unless present in the text.\
"""



"""The parameters to use when calling the Groq API."""

from __future__ import annotations

from groq.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from groq.types.chat.completion_create_params import ResponseFormat
from typing_extensions import NotRequired

from ..base import BaseCallParams


class GroqCallParams(BaseCallParams):
    """The parameters to use when calling the Groq API."""

    frequency_penalty: NotRequired[float | None]
    logit_bias: NotRequired[dict[str, int] | None]
    logprobs: NotRequired[bool | None]
    max_tokens: NotRequired[int | None]
    n: NotRequired[int | None]
    parallel_tool_calls: NotRequired[bool]
    presence_penalty: NotRequired[float | None]
    response_format: NotRequired[ResponseFormat]
    seed: NotRequired[int | None]
    stop: NotRequired[str | list[str] | None]
    temperature: NotRequired[float | None]
    tool_choice: NotRequired[ChatCompletionToolChoiceOptionParam]
    top_logprobs: NotRequired[int | None]
    top_p: NotRequired[float | None]
    user: NotRequired[str]



"""The Mirascope OpenAI Module."""

from ._call import openai_call
from ._call import openai_call as call
from .call_params import OpenAICallParams
from .call_response import OpenAICallResponse
from .call_response_chunk import OpenAICallResponseChunk
from .dynamic_config import OpenAIDynamicConfig
from .stream import OpenAIStream
from .tool import OpenAITool

__all__ = [
    "call",
    "OpenAIDynamicConfig",
    "OpenAICallParams",
    "OpenAICallResponse",
    "OpenAICallResponseChunk",
    "OpenAIStream",
    "OpenAITool",
    "openai_call",
]



"""The Mirascope Groq Module."""

from ._call import groq_call
from ._call import groq_call as call
from .call_params import GroqCallParams
from .call_response import GroqCallResponse
from .call_response_chunk import GroqCallResponseChunk
from .dynamic_config import GroqDynamicConfig
from .stream import GroqStream
from .tool import GroqTool

__all__ = [
    "call",
    "GroqDynamicConfig",
    "GroqCallParams",
    "GroqCallResponse",
    "GroqCallResponseChunk",
    "GroqStream",
    "GroqTool",
    "groq_call",
]



"""A module for interacting with OpenAI models."""

from .embedders import OpenAIEmbedder
from .embedding_params import OpenAIEmbeddingParams
from .embedding_response import OpenAIEmbeddingResponse

__all__ = [
    "OpenAIEmbedder",
    "OpenAIEmbeddingParams",
    "OpenAIEmbeddingResponse",
]



"""Calculate the cost of a completion using the Mistral API."""


def calculate_cost(
    input_tokens: int | float | None,
    output_tokens: int | float | None,
    model="open-mistral-7b",
) -> float | None:
    """Calculate the cost of a completion using the Mistral API.

    https://mistral.ai/technology/#pricing

    Model                     Input               Output
    open-mistral-nemo         $0.3/1M tokens	  $0.3/1M tokens
    mistral-large-latest      $3/1M tokens	      $9/1M tokens
    codestral-2405            $1/1M tokens	      $3/1M tokens
    open-mistral-7b	          $0.25/1M tokens	  $0.25/1M tokens
    open-mixtral-8x7b	      $0.7/1M tokens	  $0.7/1M tokens
    open-mixtral-8x22b	      $2/1M tokens	      $6/1M tokens
    mistral-small-latest	  $2/1M tokens	      $6/1M tokens
    mistral-medium-latest     $2.75/1M tokens	  $8.1/1M tokens
    """
    pricing = {
        "open-mistral-nemo": {"prompt": 0.000_000_3, "completion": 0.000_000_3},
        "open-mistral-nemo-2407": {"prompt": 0.000_000_3, "completion": 0.000_000_3},
        "mistral-large-latest": {"prompt": 0.000_003, "completion": 0.000_009},
        "mistral-large-2407": {"prompt": 0.000_003, "completion": 0.000_009},
        "open-mistral-7b": {"prompt": 0.000_000_25, "completion": 0.000_000_25},
        "open-mixtral-8x7b": {"prompt": 0.000_000_7, "completion": 0.000_000_7},
        "open-mixtral-8x22b": {"prompt": 0.000_002, "completion": 0.000_006},
        "mistral-small-latest": {"prompt": 0.000_002, "completion": 0.000_006},
        "mistral-medium-latest": {"prompt": 0.000_002_75, "completion": 0.000_008_1},
    }

    if input_tokens is None or output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_cost = input_tokens * model_pricing["prompt"]
    completion_cost = output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    return total_cost



"""The `middleware_factory` method for handling the call response."""

import inspect
from contextlib import AbstractContextManager, contextmanager
from functools import wraps
from typing import (
    Awaitable,
    Callable,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from pydantic import BaseModel

from mirascope.core.base._utils._base_type import BaseType

from ..core.base._stream import BaseStream
from ..core.base._structured_stream import BaseStructuredStream
from ..core.base.call_response import BaseCallResponse

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseStructuredStreamT = TypeVar("_BaseStructuredStreamT", bound=BaseStructuredStream)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
ResponseModel = BaseModel | BaseType
_P = ParamSpec("_P")
SyncFunc = Callable[
    _P, _BaseCallResponseT | _BaseStreamT | _ResponseModelT | _BaseStructuredStreamT
]
AsyncFunc = Callable[
    _P,
    Awaitable[
        _BaseCallResponseT | _BaseStreamT | _ResponseModelT | _BaseStructuredStreamT
    ],
]
_T = TypeVar("_T")


@contextmanager
def default_context_manager(
    fn: SyncFunc | AsyncFunc,
):
    yield None


@overload
def middleware_decorator(
    fn: SyncFunc,
    *,
    custom_context_manager: Callable[
        [SyncFunc], AbstractContextManager[_T]
    ] = default_context_manager,
    custom_decorator: Callable | None = None,
    handle_call_response: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_call_response_async: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_stream: Callable[[BaseStream, SyncFunc | AsyncFunc, _T | None], None]
    | None = None,
    handle_stream_async: Callable[
        [BaseStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_response_model: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_response_model_async: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_structured_stream: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_structured_stream_async: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
) -> SyncFunc: ...  # pragma: no cover


@overload
def middleware_decorator(
    fn: AsyncFunc,
    *,
    custom_context_manager: Callable[
        [AsyncFunc], AbstractContextManager[_T]
    ] = default_context_manager,
    custom_decorator: Callable | None = None,
    handle_call_response: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_call_response_async: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_stream: Callable[[BaseStream, SyncFunc | AsyncFunc, _T | None], None]
    | None = None,
    handle_stream_async: Callable[
        [BaseStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_response_model: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_response_model_async: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_structured_stream: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_structured_stream_async: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
) -> AsyncFunc: ...  # pragma: no cover


def middleware_decorator(
    fn: SyncFunc | AsyncFunc,
    *,
    custom_context_manager: Callable[
        [SyncFunc | AsyncFunc], AbstractContextManager[_T]
    ] = default_context_manager,
    custom_decorator: Callable | None = None,
    handle_call_response: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_call_response_async: Callable[
        [BaseCallResponse, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_stream: Callable[[BaseStream, SyncFunc | AsyncFunc, _T | None], None]
    | None = None,
    handle_stream_async: Callable[
        [BaseStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_response_model: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_response_model_async: Callable[
        [ResponseModel, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
    handle_structured_stream: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], None
    ]
    | None = None,
    handle_structured_stream_async: Callable[
        [BaseStructuredStream, SyncFunc | AsyncFunc, _T | None], Awaitable[None]
    ]
    | None = None,
) -> SyncFunc | AsyncFunc:
    is_async = inspect.iscoroutinefunction(fn)

    @wraps(fn)
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> SyncFunc:
        result = fn(*args, **kwargs)
        if isinstance(result, BaseCallResponse) and handle_call_response is not None:
            with custom_context_manager(fn) as context:
                handle_call_response(result, fn, context)
        elif isinstance(result, BaseStream):
            original_iter = result.__iter__

            def new_stream_iter(self):
                # Create the context manager when user iterates over the stream
                with custom_context_manager(fn) as context:
                    for chunk, tool in original_iter():
                        yield chunk, tool
                    if handle_stream is not None:
                        handle_stream(result, fn, context)

            result.__class__.__iter__ = (
                custom_decorator(new_stream_iter)
                if custom_decorator
                else new_stream_iter
            )
        elif isinstance(result, ResponseModel) and handle_response_model is not None:
            with custom_context_manager(fn) as context:
                handle_response_model(result, fn, context)
        elif isinstance(result, BaseStructuredStream):
            original_iter = result.__iter__

            def new_iter(self):
                # Create the context manager when user iterates over the stream
                with custom_context_manager(fn) as context:
                    for chunk in original_iter():
                        yield chunk
                    if handle_structured_stream is not None:
                        handle_structured_stream(result, fn, context)

            result.__class__.__iter__ = (
                custom_decorator(new_iter) if custom_decorator else new_iter
            )
        return cast(SyncFunc, result)

    @wraps(fn)
    async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> AsyncFunc:
        result = await fn(*args, **kwargs)
        if (
            isinstance(result, BaseCallResponse)
            and handle_call_response_async is not None
        ):
            with custom_context_manager(fn) as context:
                await handle_call_response_async(result, fn, context)
        elif isinstance(result, BaseStream):
            original_aiter = result.__aiter__

            def new_aiter_stream(self):
                async def generator():
                    with custom_context_manager(fn) as context:
                        async for chunk, tool in original_aiter():
                            yield chunk, tool
                        if handle_stream_async is not None:
                            await handle_stream_async(result, fn, context)

                return generator()

            result.__class__.__aiter__ = (
                custom_decorator(new_aiter_stream)
                if custom_decorator
                else new_aiter_stream
            )
        elif (
            isinstance(result, ResponseModel)
            and handle_response_model_async is not None
        ):
            with custom_context_manager(fn) as context:
                await handle_response_model_async(result, fn, context)
        elif isinstance(result, BaseStructuredStream):
            original_aiter = result.__aiter__

            def new_aiter(self):
                async def generator():
                    with custom_context_manager(fn) as context:
                        async for chunk in original_aiter():
                            yield chunk
                        if handle_structured_stream_async is not None:
                            await handle_structured_stream_async(result, fn, context)

                return generator()

            result.__class__.__aiter__ = (
                custom_decorator(new_aiter) if custom_decorator else new_aiter
            )
        return cast(AsyncFunc, result)

    inner_fn = inner_async if is_async else inner
    decorated_fn = custom_decorator(inner_fn) if custom_decorator else inner_fn
    return decorated_fn



"""Mirascope x Tenacity Integration."""

from pydantic import ValidationError
from tenacity import RetryCallState


def collect_validation_errors(retry_state: RetryCallState) -> None:
    """Collect errors into a `validation_errors` kwarg of the wrapped function."""
    if (
        (outcome := retry_state.outcome)
        and (exception := outcome.exception())
        and type(exception) == ValidationError
    ):
        errors = retry_state.kwargs.pop("validation_errors", [])
        retry_state.kwargs["validation_errors"] = (
            errors + [exception] if errors else [exception]
        )



"""The Mirascope Gemini Module."""

from ._call import gemini_call
from ._call import gemini_call as call
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .stream import GeminiStream
from .tool import GeminiTool

__all__ = [
    "call",
    "GeminiDynamicConfig",
    "GeminiCallParams",
    "GeminiCallResponse",
    "GeminiCallResponseChunk",
    "GeminiStream",
    "GeminiTool",
    "gemini_call",
]



"""The `GeminiStream` class for convenience around streaming LLM calls."""

from typing import cast

from google.ai.generativelanguage import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
)
from google.generativeai.types import (
    ContentDict,
    ContentsType,  # type: ignore
)
from google.generativeai.types import (  # type: ignore
    GenerateContentResponse as GenerateContentResponseType,
)
from google.generativeai.types.content_types import PartType

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .tool import GeminiTool


class GeminiStream(
    BaseStream[
        GeminiCallResponse,
        GeminiCallResponseChunk,
        ContentDict,
        ContentDict,
        ContentDict,
        ContentsType,
        GeminiTool,
        GeminiDynamicConfig,
        GeminiCallParams,
        Candidate.FinishReason,
    ]
):
    _provider = "gemini"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self, tool_calls: list[FunctionCall] | None = None, content: str | None = None
    ) -> ContentDict:
        """Constructs the message parameter for the assistant."""
        return {
            "role": "model",
            "parts": cast(list[PartType], [{"text": content}] + (tool_calls or [])),
        }

    def construct_call_response(self) -> GeminiCallResponse:
        """Constructs the call response from a consumed GeminiStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        response = GenerateContentResponseType.from_response(
            GenerateContentResponse(
                candidates=[
                    Candidate(
                        finish_reason=self.finish_reasons[0]
                        if self.finish_reasons
                        else Candidate.FinishReason.STOP,
                        content=Content(
                            role=self.message_param["role"],
                            parts=self.message_param["parts"],
                        ),
                    )
                ]
            )
        )
        return GeminiCallResponse(
            metadata=self.metadata,
            response=response,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



"""Calculate the cost of a completion using the Groq API."""


def calculate_cost(
    input_tokens: int | float | None,
    output_tokens: int | float | None,
    model="mixtral-8x7b-32768",
) -> float | None:
    """Calculate the cost of a completion using the Groq API.

    https://wow.groq.com/

    Model                                  Input               Output
    llama-3.1-405b-reasoning               N/A                 N/A
    llama-3.1-70b-versatile                N/A                 N/A
    llama-3.1-8b-instant                   N/A                 N/A
    llama3-groq-70b-8192-tool-use-preview  $0.89 / 1M tokens   $0.89 / 1M tokens
    llama3-groq-8b-8192-tool-use-preview   $0.19 / 1M tokens   $0.19 / 1M tokens
    llama3-70b-8192                        $0.59 / 1M tokens   $0.79 / 1M tokens
    llama3-8b-8192                         $0.05 / 1M tokens   $0.08 / 1M tokens
    mixtral-8x7b-32768                     $0.27 / 1M tokens   $0.27 / 1M tokens
    gemma-7b-it                            $0.07 / 1M tokens   $0.07 / 1M tokens
    gemma2-9b-it                           $0.20 / 1M tokens   $0.20 / 1M tokens
    """
    pricing = {
        "llama3-groq-70b-8192-tool-use-preview": {
            "prompt": 0.000_000_89,
            "completion": 0.000_000_89,
        },
        "llama3-groq-8b-8192-tool-use-preview": {
            "prompt": 0.000_000_19,
            "completion": 0.000_000_19,
        },
        "llama3-70b-8192": {
            "prompt": 0.000_000_59,
            "completion": 0.000_000_79,
        },
        "llama3-8b-8192": {
            "prompt": 0.000_000_05,
            "completion": 0.000_000_08,
        },
        "mixtral-8x7b-32768": {
            "prompt": 0.000_000_24,
            "completion": 0.000_000_24,
        },
        "gemma-7b-it": {
            "prompt": 0.000_000_07,
            "completion": 0.000_000_07,
        },
        "gemma2-9b-it": {
            "prompt": 0.000_000_2,
            "completion": 0.000_000_2,
        },
    }

    if input_tokens is None or output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_cost = input_tokens * model_pricing["prompt"]
    completion_cost = output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    return total_cost



"""The `OpenAITool` class for easy tool usage with OpenAI LLM calls."""

from __future__ import annotations

import jiter
from openai.types.chat import (
    ChatCompletionMessageToolCall,
    ChatCompletionToolParam,
)
from openai.types.shared_params import FunctionDefinition
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class OpenAITool(BaseTool):
    """A class for defining tools for OpenAI LLM calls."""

    tool_call: SkipJsonSchema[ChatCompletionMessageToolCall]

    @classmethod
    def tool_schema(cls) -> ChatCompletionToolParam:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined."""
        fn = FunctionDefinition(name=cls._name(), description=cls._description())
        model_schema = cls.model_tool_schema()
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        return ChatCompletionToolParam(function=fn, type="function")

    @classmethod
    def from_tool_call(cls, tool_call: ChatCompletionMessageToolCall) -> OpenAITool:
        """Constructs an `OpenAITool` instance from a `tool_call`."""
        model_json = jiter.from_json(tool_call.function.arguments.encode())
        model_json["tool_call"] = tool_call.model_dump()
        return cls.model_validate(model_json)



"""The `AnthropicStream` class for convenience around streaming LLM calls."""

from anthropic.types import Message, MessageParam, TextBlock, ToolUseBlock, Usage
from anthropic.types.content_block import ContentBlock
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_use_block_param import ToolUseBlockParam
from pydantic import BaseModel

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import AnthropicCallParams
from .call_response import AnthropicCallResponse
from .call_response_chunk import AnthropicCallResponseChunk
from .dynamic_config import AnthropicDynamicConfig
from .tool import AnthropicTool

FinishReason = Message.__annotations__["stop_reason"]


class AnthropicStream(
    BaseStream[
        AnthropicCallResponse,
        AnthropicCallResponseChunk,
        MessageParam,
        MessageParam,
        MessageParam,
        MessageParam,
        AnthropicTool,
        AnthropicDynamicConfig,
        AnthropicCallParams,
        FinishReason,
    ]
):
    _provider = "anthropic"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self, tool_calls: list[ToolUseBlock] | None = None, content: str | None = None
    ) -> MessageParam:
        """Constructs the message parameter for the assistant."""
        message_param_tool_calls = []
        if tool_calls:
            message_param_tool_calls: list[ToolUseBlockParam] = [
                ToolUseBlockParam(
                    id=tool_call.id,
                    input=tool_call.input,
                    name=tool_call.name,
                    type="tool_use",
                )
                for tool_call in tool_calls
            ]
        return {
            "role": "assistant",
            "content": ([TextBlockParam(text=content, type="text")] if content else [])
            + message_param_tool_calls,
        }

    def construct_call_response(self) -> AnthropicCallResponse:
        """Constructs the call response from a consumed AnthropicStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        usage = Usage(
            input_tokens=int(self.input_tokens or 0),
            output_tokens=int(self.output_tokens or 0),
        )

        content_blocks: list[ContentBlock] = []

        if isinstance(self.message_param["content"], str):
            content_blocks.append(
                TextBlock(text=self.message_param["content"], type="text")
            )
        else:
            for content in self.message_param["content"]:
                content_type = (
                    content.type if isinstance(content, BaseModel) else content["type"]
                )

                if content_type == "text":
                    content_blocks.append(TextBlock.model_validate(content))
                elif content_type == "tool_use":
                    content_blocks.append(ToolUseBlock.model_validate(content))
        completion = Message(
            id=self.id if self.id else "",
            content=content_blocks,
            model=self.model,
            role="assistant",
            stop_reason=self.finish_reasons[0] if self.finish_reasons else None,
            stop_sequence=None,
            type="message",
            usage=usage,
        )
        return AnthropicCallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



"""This module contains the base call response class."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, TypeVar

from pydantic import (
    BaseModel,
    ConfigDict,
    SkipValidation,
    computed_field,
    field_serializer,
)

from .call_params import BaseCallParams
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata
from .tool import BaseTool

_ResponseT = TypeVar("_ResponseT", bound=Any)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_MessageParamT = TypeVar("_MessageParamT", bound=Any)
_CallParamsT = TypeVar("_CallParamsT", bound=BaseCallParams)
_UserMessageParamT = TypeVar("_UserMessageParamT", bound=Any)


class BaseCallResponse(
    BaseModel,
    Generic[
        _ResponseT,
        _BaseToolT,
        _BaseDynamicConfigT,
        _MessageParamT,
        _CallParamsT,
        _UserMessageParamT,
    ],
    ABC,
):
    """A base abstract interface for LLM call responses.

    Attributes:
        response: The original response from whichever model response this wraps.
        user_message_param: The most recent message if it was a user message. Otherwise
            `None`.
        tool_types: The tool types sent in the LLM call.
        start_time: The start time of the completion in ms.
        end_time: The end time of the completion in ms.
        cost: The cost of the completion in dollars.
    """

    metadata: Metadata
    response: _ResponseT
    tool_types: list[type[_BaseToolT]] | None = None
    prompt_template: str | None
    fn_args: dict[str, Any]
    dynamic_config: _BaseDynamicConfigT
    messages: SkipValidation[list[_MessageParamT]]
    call_params: SkipValidation[_CallParamsT]
    call_kwargs: dict[str, Any]
    user_message_param: _UserMessageParamT | None = None
    start_time: float
    end_time: float

    _provider: ClassVar[str] = "NO PROVIDER"
    _model: str = "NO MODEL"

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @field_serializer("tool_types")
    def serialize_tool_types(self, tool_types: list[type[_BaseToolT]], _info):
        return [{"type": "function", "name": tool._name()} for tool in tool_types or []]

    def __str__(self) -> str:
        """Returns the string content of the response."""
        return self.content

    @property
    @abstractmethod
    def content(self) -> str:
        """Should return the string content of the response.

        If there are multiple choices in a response, this method should select the 0th
        choice and return it's string content.

        If there is no string content (e.g. when using tools), this method must return
        the empty string.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def finish_reasons(self) -> list[str] | None:
        """Should return the finish reasons of the response.

        If there is no finish reason, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def model(self) -> str | None:
        """Should return the name of the response model."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def id(self) -> str | None:
        """Should return the id of the response."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def usage(self) -> Any:
        """Should return the usage of the response.

        If there is no usage, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def input_tokens(self) -> int | float | None:
        """Should return the number of input tokens.

        If there is no input_tokens, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def output_tokens(self) -> int | float | None:
        """Should return the number of output tokens.

        If there is no output_tokens, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def cost(self) -> float | None:
        """Should return the cost of the response in dollars.

        If there is no cost, this method must return None.
        """
        ...  # pragma: no cover

    @computed_field
    @property
    @abstractmethod
    def message_param(self) -> Any:
        """Returns the assistant's response as a message parameter."""
        ...  # pragma: no cover

    @computed_field
    @property
    @abstractmethod
    def tools(self) -> list[_BaseToolT] | None:
        """Returns the tools for the 0th choice message."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def tool(self) -> _BaseToolT | None:
        """Returns the 0th tool for the 0th choice message."""
        ...  # pragma: no cover

    @classmethod
    @abstractmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[_BaseToolT, Any]]
    ) -> list[Any]:
        """Returns the tool message parameters for tool call results."""
        ...  # pragma: no cover



"""The Mirascope Core Functionality."""

from contextlib import suppress

from . import base
from .base import (
    BasePrompt,
    BaseTool,
    BaseToolKit,
    metadata,
    prompt_template,
    toolkit_tool,
)

with suppress(ImportError):
    from . import anthropic as anthropic

with suppress(ImportError):
    from . import cohere as cohere

with suppress(ImportError):
    from . import gemini as gemini

with suppress(ImportError):
    from . import groq as groq


with suppress(ImportError):
    from . import litellm as litellm


with suppress(ImportError):
    from . import mistral as mistral

with suppress(ImportError):
    from . import openai as openai

__all__ = [
    "anthropic",
    "base",
    "BasePrompt",
    "BaseTool",
    "BaseToolKit",
    "cohere",
    "gemini",
    "groq",
    "litellm",
    "metadata",
    "mistral",
    "openai",
    "prompt_template",
    "toolkit_tool",
]



from typing import Callable

from langfuse.decorators import langfuse_context
from pydantic import BaseModel

from mirascope.core.base._utils._base_type import BaseType

from ...core.base import BaseCallResponse
from ...core.base._stream import BaseStream
from ...core.base._structured_stream import BaseStructuredStream
from ...core.base._utils import get_metadata


class ModelUsage(BaseModel):
    input: int | float | None
    output: int | float | None
    unit: str


def get_call_response_observation(result: BaseCallResponse, fn: Callable):
    metadata = get_metadata(fn, {})
    tags = metadata.get("tags", [])
    return {
        "name": f"{fn.__name__} with {result.model}",
        "input": result.prompt_template,
        "metadata": result.response,
        "tags": tags,
        "model": result.model,
        "output": result.message_param.get("content", None),
    }


def handle_call_response(result: BaseCallResponse, fn: Callable, context: None):
    langfuse_context.update_current_observation(
        **get_call_response_observation(result, fn),
        usage=ModelUsage(
            input=result.input_tokens, output=result.output_tokens, unit="TOKENS"
        ),
    )


def handle_stream(stream: BaseStream, fn: Callable, context: None):
    usage = ModelUsage(
        input=stream.input_tokens,
        output=stream.output_tokens,
        unit="TOKENS",
    )
    langfuse_context.update_current_observation(
        **get_call_response_observation(stream.construct_call_response(), fn),
        usage=usage,
    )


def handle_response_model(result: BaseModel | BaseType, fn: Callable, context: None):
    if isinstance(result, BaseModel):
        response: BaseCallResponse = result._response  # type: ignore
        call_response_observation = get_call_response_observation(response, fn)
        call_response_observation.pop("output")
        langfuse_context.update_current_observation(
            **call_response_observation,
            usage=ModelUsage(
                input=response.input_tokens,
                output=response.output_tokens,
                unit="TOKENS",
            ),
            output=result,
        )
    else:
        langfuse_context.update_current_observation(
            output=result,
        )


def handle_structured_stream(result: BaseStructuredStream, fn: Callable, context: None):
    stream: BaseStream = result.stream
    usage = ModelUsage(
        input=stream.input_tokens,
        output=stream.output_tokens,
        unit="TOKENS",
    )
    langfuse_context.update_current_observation(
        **get_call_response_observation(stream.construct_call_response(), fn),
        usage=usage,
        output=result.constructed_response_model,
    )


async def handle_call_response_async(
    result: BaseCallResponse, fn: Callable, context: None
):
    handle_call_response(result, fn, None)


async def handle_stream_async(stream: BaseStream, fn: Callable, context: None):
    handle_stream(stream, fn, None)


async def handle_response_model_async(
    result: BaseModel | BaseType, fn: Callable, context: None
):
    handle_response_model(result, fn, None)


async def handle_structured_stream_async(
    result: BaseStructuredStream, fn: Callable, context: None
):
    handle_structured_stream(result, fn, None)



"""Get the JSON output from a completion response."""

from ..call_response import MistralCallResponse
from ..call_response_chunk import MistralCallResponseChunk


def get_json_output(
    response: MistralCallResponse | MistralCallResponseChunk, json_mode: bool
) -> str:
    """Get the JSON output from a completion response."""
    if isinstance(response, MistralCallResponse):
        if json_mode and response.content:
            return response.content
        elif tool_calls := response.response.choices[0].message.tool_calls:
            return tool_calls[0].function.arguments
        raise ValueError("No tool call or JSON object found in response.")
    else:
        # raise ValueError("Mistral does not support structured streaming... :(")
        if json_mode:
            return response.content
        elif (
            (choices := response.chunk.choices)
            and (tool_calls := choices[0].delta.tool_calls)
            and (function := tool_calls[0].function)
            and (arguments := function.arguments) is not None
        ):
            return arguments
        return ""



"""Mirascope Base Classes."""

from . import _partial, _utils
from ._call_factory import call_factory
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .message_param import AudioPart, BaseMessageParam, ImagePart, TextPart
from .metadata import Metadata
from .prompt import BasePrompt, metadata, prompt_template
from .tool import BaseTool
from .toolkit import BaseToolKit, toolkit_tool

__all__ = [
    "AudioPart",
    "BaseCallParams",
    "BaseCallResponse",
    "BaseCallResponseChunk",
    "BaseDynamicConfig",
    "BaseMessageParam",
    "BasePrompt",
    "BaseTool",
    "BaseToolKit",
    "call_factory",
    "ImagePart",
    "metadata",
    "Metadata",
    "prompt_template",
    "TextPart",
    "toolkit_tool",
    "_partial",
    "_utils",
]



"""Mirascope x OpenTelemetry Integration utils"""

import json
from contextlib import contextmanager
from typing import Any, Callable, Generator, Sequence

from opentelemetry.sdk.trace import SpanProcessor, TracerProvider
from opentelemetry.sdk.trace.export import (
    ConsoleSpanExporter,
    SimpleSpanProcessor,
)
from opentelemetry.trace import (
    Tracer,
    get_tracer,
    set_tracer_provider,
)
from opentelemetry.trace.span import Span
from opentelemetry.util.types import Attributes, AttributeValue
from pydantic import BaseModel

from mirascope.core.base._utils._base_type import BaseType

from ...core.base import BaseCallResponse
from ...core.base._stream import BaseStream
from ...core.base._structured_stream import BaseStructuredStream


def configure(
    processors: Sequence[SpanProcessor] | None = None,
) -> Tracer:
    """Configures the OpenTelemetry tracer, this function should only be called once.

    Args:
        processors: Optional[Sequence[SpanProcessor]]
            The span processors to use, if None, a console exporter will be used.

    Returns:
        The configured tracer.
    """
    provider = TracerProvider()
    if processors is None:
        provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))
    else:
        for processor in processors:
            provider.add_span_processor(processor)
    # NOTE: Sets the global trace provider, should only be called once
    set_tracer_provider(provider)
    return get_tracer("otel")


@contextmanager
def custom_context_manager(
    fn: Callable,
) -> Generator[Span, Any, None]:
    tracer = get_tracer("otel")
    with tracer.start_as_current_span(f"{fn.__name__}") as span:
        yield span


def get_call_response_attributes(result: BaseCallResponse) -> dict[str, AttributeValue]:
    max_tokens = getattr(result.call_params, "max_tokens", 0)
    temperature = getattr(result.call_params, "temperature", 0)
    top_p = getattr(result.call_params, "top_p", 0)

    return {
        "gen_ai.system": result.prompt_template if result.prompt_template else "",
        "gen_ai.request.model": result.call_kwargs.get("model", ""),
        "gen_ai.request.max_tokens": max_tokens,
        "gen_ai.request.temperature": temperature,
        "gen_ai.request.top_p": top_p,
        "gen_ai.response.model": result.model if result.model else "",
        "gen_ai.response.id": result.id if result.id else "",
        "gen_ai.response.finish_reasons": result.finish_reasons
        if result.finish_reasons
        else "",
        "gen_ai.usage.completion_tokens": result.output_tokens
        if result.output_tokens
        else "",
        "gen_ai.usage.prompt_tokens": result.input_tokens
        if result.input_tokens
        else "",
    }


def set_call_response_event_attributes(result: BaseCallResponse, span: Span) -> None:
    prompt_attributes = {
        "gen_ai.prompt": json.dumps(result.user_message_param),
    }
    completion_attributes: Attributes = {
        "gen_ai.completion": json.dumps(result.message_param),
    }
    span.add_event(
        "gen_ai.content.prompt",
        attributes=prompt_attributes,
    )
    span.add_event(
        "gen_ai.content.completion",
        attributes=completion_attributes,
    )


def handle_call_response(result: BaseCallResponse, fn: Callable, span: Span | None):
    if span is None:
        return

    attributes = get_call_response_attributes(result)
    attributes["async"] = False
    span.set_attributes(attributes)
    set_call_response_event_attributes(result, span)


def handle_stream(stream: BaseStream, fn: Callable, span: Span | None):
    if span is None:
        return
    constructed_call_response = stream.construct_call_response()
    attributes = get_call_response_attributes(constructed_call_response)
    attributes["async"] = False
    span.set_attributes(attributes)
    set_call_response_event_attributes(constructed_call_response, span)


def handle_response_model(
    result: BaseModel | BaseType, fn: Callable, span: Span | None
):
    if span is None:
        return
    if isinstance(result, BaseModel):
        response: BaseCallResponse = result._response  # type: ignore
        attributes = get_call_response_attributes(response)
        attributes["async"] = False
        span.set_attributes(attributes)
        prompt_attributes = {
            "gen_ai.prompt": json.dumps(response.user_message_param),
        }
        completion_attributes: Attributes = {
            "gen_ai.completion": result.model_dump_json(),
        }
        span.add_event(
            "gen_ai.content.prompt",
            attributes=prompt_attributes,
        )
    else:
        span.set_attributes({"async": False})
        if not isinstance(result, str | int | float | bool):
            result = str(result)
        completion_attributes: Attributes = {
            "gen_ai.completion": result,
        }
    span.add_event(
        "gen_ai.content.completion",
        attributes=completion_attributes,
    )


def handle_structured_stream(
    result: BaseStructuredStream, fn: Callable, span: Span | None
):
    if span is None:
        return
    attributes = get_call_response_attributes(result.stream.construct_call_response())
    attributes["async"] = False
    span.set_attributes(attributes)
    prompt_attributes = {
        "gen_ai.prompt": json.dumps(result.stream.user_message_param),
    }
    if isinstance(result.constructed_response_model, BaseModel):
        completion = result.constructed_response_model.model_dump_json()
    else:
        completion = result.constructed_response_model
    completion_attributes: Attributes = {
        "gen_ai.completion": completion,
    }
    span.add_event(
        "gen_ai.content.prompt",
        attributes=prompt_attributes,
    )
    span.add_event(
        "gen_ai.content.completion",
        attributes=completion_attributes,
    )


async def handle_call_response_async(
    result: BaseCallResponse, fn: Callable, span: Span | None
):
    if span is None:
        return

    attributes = get_call_response_attributes(result)
    attributes["async"] = True
    span.set_attributes(attributes)
    set_call_response_event_attributes(result, span)


async def handle_stream_async(stream: BaseStream, fn: Callable, span: Span | None):
    if span is None:
        return
    constructed_call_response = stream.construct_call_response()
    attributes = get_call_response_attributes(constructed_call_response)
    attributes["async"] = True
    span.set_attributes(attributes)
    set_call_response_event_attributes(constructed_call_response, span)


async def handle_response_model_async(
    result: BaseModel | BaseType, fn: Callable, span: Span | None
):
    if span is None:
        return
    if isinstance(result, BaseModel):
        response: BaseCallResponse = result._response  # type: ignore
        attributes = get_call_response_attributes(response)
        attributes["async"] = True
        span.set_attributes(attributes)
        prompt_attributes = {
            "gen_ai.prompt": json.dumps(response.user_message_param),
        }
        completion_attributes: Attributes = {
            "gen_ai.completion": result.model_dump_json(),
        }
        span.add_event(
            "gen_ai.content.prompt",
            attributes=prompt_attributes,
        )
    else:
        span.set_attributes({"async": True})
        if not isinstance(result, str | int | float | bool):
            result = str(result)
        completion_attributes: Attributes = {
            "gen_ai.completion": result,
        }
    span.add_event(
        "gen_ai.content.completion",
        attributes=completion_attributes,
    )


async def handle_structured_stream_async(
    result: BaseStructuredStream, fn: Callable, span: Span | None
):
    if span is None:
        return
    attributes = get_call_response_attributes(result.stream.construct_call_response())
    attributes["async"] = True
    span.set_attributes(attributes)
    prompt_attributes = {
        "gen_ai.prompt": json.dumps(result.stream.user_message_param),
    }
    if isinstance(result.constructed_response_model, BaseModel):
        completion = result.constructed_response_model.model_dump_json()
    else:
        completion = result.constructed_response_model
    completion_attributes: Attributes = {
        "gen_ai.completion": completion,
    }
    span.add_event(
        "gen_ai.content.prompt",
        attributes=prompt_attributes,
    )
    span.add_event(
        "gen_ai.content.completion",
        attributes=completion_attributes,
    )



from typing import Literal

from cohere.types import EmbedByTypeResponseEmbeddings, EmbedResponse
from pydantic import SkipValidation

from ..base.embedding_response import BaseEmbeddingResponse


class CohereEmbeddingResponse(BaseEmbeddingResponse[SkipValidation[EmbedResponse]]):
    """A convenience wrapper around the Cohere `EmbedResponse` response."""

    embedding_type: Literal["float", "int8", "uint8", "binary", "ubinary"] | None = None

    @property
    def embeddings(
        self,
    ) -> list[list[float]] | list[list[int]] | None:
        """Returns the embeddings"""
        if self.response.response_type == "embeddings_floats":
            return self.response.embeddings
        else:
            embedding_type = self.embedding_type
            if embedding_type == "float":
                embedding_type == "float_"

            # TODO: Update to model_dump when Cohere updates to Pydantic v2
            embeddings_by_type: EmbedByTypeResponseEmbeddings = self.response.embeddings
            embedding_dict = embeddings_by_type.dict()
            return embedding_dict.get(str(embedding_type), None)



"""Utility for converting `BaseMessageParam` to `MessageParam`"""

import base64

from anthropic.types import MessageParam

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | MessageParam],
) -> list[MessageParam]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(message_param.model_dump())
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.model_dump())

                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/gif",
                        "image/webp",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. Anthropic"
                            " currently only supports JPEG, PNG, GIF, and WebP images."
                        )
                    converted_content.append(
                        {
                            "type": "image",
                            "source": {
                                "data": base64.b64encode(part.image).decode("utf-8"),
                                "media_type": part.media_type,
                                "type": "base64",
                            },
                        }
                    )
                else:
                    raise ValueError(
                        "Anthropic currently only supports text and image modalities. "
                        f"Modality provided: {part.type}"
                    )
            converted_message_params.append(
                {"role": message_param.role, "content": converted_content}
            )
    return converted_message_params



"""Protocols for reusable type hints."""

import inspect
from collections.abc import AsyncGenerator, Generator
from typing import (
    Any,
    Awaitable,
    Callable,
    Literal,
    ParamSpec,
    Protocol,
    TypeGuard,
    TypeVar,
    overload,
)

from ..call_response_chunk import BaseCallResponseChunk
from ..tool import BaseTool

_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", covariant=True, bound=BaseCallResponseChunk
)
_BaseClientT = TypeVar("_BaseClientT", contravariant=True)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", contravariant=True)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", contravariant=True)
_ResponseT = TypeVar("_ResponseT", covariant=True)
_AsyncResponseT = TypeVar("_AsyncResponseT", covariant=True)
_ResponseChunkT = TypeVar("_ResponseChunkT", covariant=True)
_InvariantResponseChunkT = TypeVar("_InvariantResponseChunkT", contravariant=True)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_P = ParamSpec("_P")
_R = TypeVar("_R", contravariant=True)


class LLMFunctionDecorator(Protocol[_BaseDynamicConfigT, _ResponseT, _AsyncResponseT]):
    @overload
    def __call__(
        self, fn: Callable[_P, _BaseDynamicConfigT]
    ) -> Callable[_P, _ResponseT]: ...  # pragma: no cover

    @overload
    def __call__(
        self, fn: Callable[_P, Awaitable[_BaseDynamicConfigT]]
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...  # pragma: no cover

    def __call__(
        self, fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]]
    ) -> Callable[_P, _ResponseT | Awaitable[_AsyncResponseT]]: ...  # pragma: no cover


class AsyncCreateFn(Protocol[_ResponseT, _ResponseChunkT]):
    @overload
    def __call__(
        self, *, stream: Literal[False] = False, **kwargs: Any
    ) -> Awaitable[_ResponseT]: ...  # pragma: no cover

    @overload
    def __call__(
        self, *, stream: Literal[True] = True, **kwargs: Any
    ) -> Awaitable[AsyncGenerator[_ResponseChunkT, None]]: ...  # pragma: no cover

    def __call__(
        self, *, stream: bool = False, **kwargs: Any
    ) -> Awaitable[
        _ResponseT | AsyncGenerator[_ResponseChunkT, None]
    ]: ...  # pragma: no cover


class CreateFn(Protocol[_ResponseT, _ResponseChunkT]):
    @overload
    def __call__(
        self, *, stream: Literal[False] = False, **kwargs: Any
    ) -> _ResponseT: ...  # pragma: no cover

    @overload
    def __call__(
        self, *, stream: Literal[True] = True, **kwargs: Any
    ) -> Generator[_ResponseChunkT, None, None]: ...  # pragma: no cover

    def __call__(
        self, *, stream: bool = False, **kwargs: Any
    ) -> _ResponseT | Generator[_ResponseChunkT, None, None]: ...  # pragma: no cover


class SetupCall(
    Protocol[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ]
):
    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _BaseClientT | None,
        fn: Callable[..., _BaseDynamicConfigT],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        extract: bool,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT],
        str,
        list[dict[str, Any]],
        list[type[_BaseToolT]] | None,
        dict[str, Any],
    ]: ...  # pragma: no cover

    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _BaseClientT | None,
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        extract: bool,
    ) -> tuple[
        AsyncCreateFn[_ResponseT, _ResponseChunkT],
        str,
        list[dict[str, Any]],
        list[type[_BaseToolT]] | None,
        dict[str, Any],
    ]: ...  # pragma: no cover

    def __call__(
        self,
        *,
        model: str,
        client: _BaseClientT | None,
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        extract: bool,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT]
        | AsyncCreateFn[_ResponseT, _ResponseChunkT],
        str,
        list[dict[str, Any]],
        list[type[_BaseToolT]] | None,
        dict[str, Any],
    ]: ...  # pragma: no cover

    @staticmethod
    def fn_is_sync(
        fn: Callable[_P, _R | Awaitable[_R]],
    ) -> TypeGuard[Callable[_P, _R]]:
        return not inspect.iscoroutinefunction(fn)

    @staticmethod
    def fn_is_async(
        fn: Callable[_P, _R | Awaitable[_R]],
    ) -> TypeGuard[Callable[_P, Awaitable[_R]]]:
        return inspect.iscoroutinefunction(fn)


class HandleStream(
    Protocol[_InvariantResponseChunkT, _BaseCallResponseChunkT, _BaseToolT]
):
    def __call__(
        self,
        stream: Generator[_InvariantResponseChunkT, None, None],
        tool_types: list[type[_BaseToolT]] | None,
    ) -> Generator[
        tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None
    ]: ...  # pragma: no cover


class HandleStreamAsync(
    Protocol[_InvariantResponseChunkT, _BaseCallResponseChunkT, _BaseToolT]
):
    def __call__(
        self,
        stream: AsyncGenerator[_InvariantResponseChunkT, None],
        tool_types: list[type[_BaseToolT]] | None,
    ) -> AsyncGenerator[
        tuple[_BaseCallResponseChunkT, _BaseToolT | None], None
    ]: ...  # pragma: no cover


class GetJsonOutput(Protocol[_R]):
    def __call__(self, response: _R, json_mode: bool) -> str: ...  # pragma: no cover


class CalculateCost(Protocol):
    def __call__(
        self,
        input_tokens: int | float | None,
        output_tokens: int | float | None,
        model: str,
    ) -> float | None: ...  # pragma: no cover



"""This module contains the base class for message parameters."""

from typing import Literal

from pydantic import BaseModel


class TextPart(BaseModel):
    type: Literal["text"]
    text: str


class ImagePart(BaseModel):
    type: Literal["image"]
    media_type: str
    image: bytes
    detail: str | None


class AudioPart(BaseModel):
    type: Literal["audio"]
    media_type: str
    audio: bytes


class BaseMessageParam(BaseModel):
    """A base class for message parameters."""

    role: str
    content: str | list[TextPart | ImagePart | AudioPart]



"""This module defines the base class for tools used in LLM calls."""

from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import Any, Callable, ClassVar, TypeVar, cast

from pydantic import BaseModel, ConfigDict

from . import _utils

_BaseToolT = TypeVar("_BaseToolT")


class BaseTool(BaseModel):
    """A class for defining tools for LLM calls."""

    __custom_name__: ClassVar[str] = ""
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def _name(cls) -> str:
        """Returns the name of the tool."""
        return cls.__custom_name__ if cls.__custom_name__ else cls.__name__

    @classmethod
    def _description(cls) -> str:
        """Returns the description of the tool."""
        return (
            inspect.cleandoc(cls.__doc__)
            if cls.__doc__
            else _utils.DEFAULT_TOOL_DOCSTRING
        )

    @property
    def args(self) -> dict[str, Any]:
        """The arguments of the tool as a dictionary."""
        return {
            field: getattr(self, field)
            for field in self.model_fields
            if field != "tool_call"
        }

    @abstractmethod
    def call(self) -> Any:
        """The method to call the tool."""
        ...  # pragma: no cover

    @classmethod
    def model_tool_schema(cls) -> dict[str, Any]:
        """Returns the model_json_schema modified for reduced token usage."""
        model_schema = cls.model_json_schema()
        model_schema.pop("title", None)
        model_schema.pop("description", None)

        def remove_schema_titles(obj):
            if isinstance(obj, dict):
                # Remove the 'title' key only if it's a direct child of a schema object
                if "type" in obj or "$ref" in obj or "properties" in obj:
                    obj.pop("title", None)

                # Recursively process nested objects
                for key, value in list(obj.items()):
                    obj[key] = remove_schema_titles(value)
            elif isinstance(obj, list):
                # Recursively process list items
                return [remove_schema_titles(item) for item in obj]

            return obj

        return cast(dict[str, Any], remove_schema_titles(model_schema))

    @classmethod
    def type_from_fn(cls: _BaseToolT, fn: Callable) -> _BaseToolT:
        """Returns this tool type converted from a function."""
        return _utils.convert_function_to_base_tool(fn, cls)  # type: ignore

    @classmethod
    def type_from_base_model_type(
        cls: _BaseToolT, tool_type: type[BaseModel]
    ) -> _BaseToolT:
        """Returns this tool type converted from a given base tool type."""
        return _utils.convert_base_model_to_base_tool(tool_type, cls)  # type: ignore

    @classmethod
    def type_from_base_type(
        cls: _BaseToolT, schema: type[_utils.BaseType]
    ) -> _BaseToolT:
        """Returns this tool type converted from a base type."""
        return _utils.convert_base_type_to_base_tool(schema, cls)  # type: ignore



"""Handles the stream of completion chunks."""

from collections.abc import AsyncGenerator, Generator

from cohere.types import StreamedChatResponse

from ..call_response_chunk import CohereCallResponseChunk
from ..tool import CohereTool


def handle_stream(
    stream: Generator[StreamedChatResponse, None, None],
    tool_types: list[type[CohereTool]] | None = None,
) -> Generator[tuple[CohereCallResponseChunk, None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed.

    Note: cohere does not currently support streaming tools.
    """
    for chunk in stream:
        yield CohereCallResponseChunk(chunk=chunk), None


async def handle_stream_async(
    stream: AsyncGenerator[StreamedChatResponse, None],
    tool_types: list[type[CohereTool]] | None = None,
) -> AsyncGenerator[tuple[CohereCallResponseChunk, None], None]:
    """
    Async iterator over the stream and constructs tools as they are streamed.

    Note: cohere does not currently support streaming tools.
    """
    async for chunk in stream:
        yield CohereCallResponseChunk(chunk=chunk), None



"""The parameters to use when calling the Anthropic API."""

from __future__ import annotations

from anthropic.types.completion_create_params import Metadata
from anthropic.types.message_create_params import ToolChoice
from httpx import Timeout
from typing_extensions import NotRequired, Required

from ..base import BaseCallParams


class AnthropicCallParams(BaseCallParams):
    """The parameters to use when calling the Anthropic API."""

    max_tokens: Required[int]
    tool_choice: NotRequired[ToolChoice | None]
    metadata: NotRequired[Metadata | None]
    stop_sequences: NotRequired[list[str] | None]
    system: NotRequired[str | None]
    temperature: NotRequired[float | None]
    top_k: NotRequired[int | None]
    top_p: NotRequired[float | None]
    timeout: NotRequired[float | Timeout | None]



"""This module contains the `AnthropicCallResponseChunk` class."""

from anthropic.types import (
    Message,
    MessageDeltaUsage,
    MessageStartEvent,
    MessageStreamEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    Usage,
)

from ..base import BaseCallResponseChunk

FinishReason = Message.__annotations__["stop_reason"]


class AnthropicCallResponseChunk(
    BaseCallResponseChunk[MessageStreamEvent, FinishReason]
):
    '''A convenience wrapper around the Anthropic `ChatCompletionChunk` streamed chunks.

    When calling the Anthropic API using a function decorated with `anthropic_call` and
    `stream` set to `True`, the stream will contain `AnthropicResponseChunk` instances
    with properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.anthropic import anthropic_call

    @anthropic_call(model="claude-3-5-sonnet-20240620", stream=True)
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    stream = recommend_book("fantasy")  # response is an `AnthropicStream`
    for chunk in stream:
        print(chunk.content, end="", flush=True)
    #> Sure! I would recommend...
    ```
    '''

    @property
    def content(self) -> str:
        """Returns the string content of the 0th message."""
        return (
            self.chunk.delta.text
            if self.chunk.type == "content_block_delta"
            and self.chunk.delta.type == "text_delta"
            else ""
        )

    @property
    def finish_reasons(self) -> list[FinishReason] | None:
        """Returns the finish reason of the response."""
        if (
            isinstance(self.chunk, RawMessageStartEvent)
            and self.chunk.message.stop_reason
        ):
            return [str(self.chunk.message.stop_reason)]
        return None

    @property
    def model(self) -> str | None:
        """Returns the name of the response model."""
        if isinstance(self.chunk, MessageStartEvent):
            return self.chunk.message.model
        return None

    @property
    def id(self) -> str | None:
        """Returns the id of the response."""
        if isinstance(self.chunk, MessageStartEvent):
            return self.chunk.message.id
        return None

    @property
    def usage(self) -> Usage | MessageDeltaUsage | None:
        """Returns the usage of the message."""
        if isinstance(self.chunk, MessageStartEvent):
            return self.chunk.message.usage
        elif isinstance(self.chunk, RawMessageDeltaEvent):
            return self.chunk.usage
        return None

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        if (usage := self.usage) and hasattr(usage, "input_tokens"):
            return getattr(usage, "input_tokens")
        return None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.output_tokens
        return None



from functools import cached_property
from typing import Any, ClassVar, Union

import weaviate
import weaviate.classes as wvc
from weaviate import WeaviateClient
from weaviate.collections.collection import Collection

from ..base.document import Document
from ..base.vectorstores import BaseVectorStore
from .types import (
    WeaviateParams,
    WeaviateQueryResult,
    WeaviateSettings,
)


class WeaviateVectorStore(BaseVectorStore):
    """A vectorstore for Weaviate.

    Example:

    ```python
    from mirascope.beta.rag.weaviate.types import WeaviateSettings, WeaviateParams
    from mirascope.beta.rag.weaviate import WeaviateVectorStore
    from mirascope.beta.rag import TextChunker


    class MyStore(WeaviateVectorStore):
        chunker = TextChunker(chunk_size=1000, chunk_overlap=200)
        index_name = "my-store-0001"
        client_settings = WeaviateSettings()

    my_store = MyStore()
    with open(f"{PATH_TO_FILE}") as file:
        data = file.read()
        my_store.add(data)
    properties = my_store.retrieve("my question").properties
    print(documents)
    """

    _provider: ClassVar[str] = "weaviate"
    vectorstore_params = WeaviateParams()
    client_settings: ClassVar[WeaviateSettings] = WeaviateSettings()

    def add(self, text: Union[str, list[Document]], **kwargs: Any) -> None:
        """Takes unstructured data and inserts into vectorstore"""
        documents: list[Document]
        if isinstance(text, str):
            chunk = self.chunker.chunk
            documents = chunk(text)
        else:
            documents = text

        if len(documents) < 2:
            return self._index.data.insert(
                properties={"text": documents[0].text}, uuid=documents[0].id, **kwargs
            )

        data_objects = list()
        for document in documents:
            data_object = wvc.data.DataObject(
                properties={"text": document.text}, uuid=document.id
            )
            data_objects.append(data_object)

        return self._index.data.insert_many(data_objects)

    def retrieve(self, text: str, **kwargs: Any) -> WeaviateQueryResult:
        """Queries the vectorstore for closest match"""
        query_result = self._index.query.near_text(query=text, **kwargs)
        result = query_result.objects[0]

        return WeaviateQueryResult.from_response(result)

    def close_connection(self):
        self._client.close()

    ############################# PRIVATE PROPERTIES #################################

    @cached_property
    def _client(self) -> WeaviateClient:
        if self.client_settings.mode == "local":
            return weaviate.connect_to_local(**self.client_settings.kwargs())
        elif self.client_settings.mode == "embedded":
            return weaviate.connect_to_embedded(**self.client_settings.kwargs())
        elif self.client_settings.mode == "cloud":
            return weaviate.connect_to_weaviate_cloud(**self.client_settings.kwargs())
        elif self.client_settings.mode == "custom":
            return weaviate.connect_to_custom(**self.client_settings.kwargs())

    @cached_property
    def _index(self) -> Collection:
        vectorstore_params = self.vectorstore_params

        if self.index_name:
            vectorstore_params = self.vectorstore_params.model_copy(
                update={"name": self.index_name}
            )
        if self._client.collections.exists(self.index_name):
            return self._client.collections.get(self.index_name)

        return self._client.collections.create(**vectorstore_params.kwargs())



"""The `CohereStream` class for convenience around streaming LLM calls."""

from cohere.types import (
    ApiMeta,
    ApiMetaBilledUnits,
    ChatMessage,
    ChatStreamEndEventFinishReason,
    NonStreamedChatResponse,
    ToolCall,
)

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import CohereCallParams
from .call_response import CohereCallResponse
from .call_response_chunk import CohereCallResponseChunk
from .dynamic_config import CohereDynamicConfig
from .tool import CohereTool


class CohereStream(
    BaseStream[
        CohereCallResponse,
        CohereCallResponseChunk,
        ChatMessage,
        ChatMessage,
        ChatMessage,
        ChatMessage,
        CohereTool,
        CohereDynamicConfig,
        CohereCallParams,
        ChatStreamEndEventFinishReason,
    ]
):
    _provider = "cohere"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self, tool_calls: list[ToolCall] | None = None, content: str | None = None
    ) -> ChatMessage:
        return ChatMessage(
            role="assistant",  # type: ignore
            message=content if content else "",
            tool_calls=tool_calls,
        )

    def construct_call_response(self) -> CohereCallResponse:
        """Constructs the call response from a consumed CohereStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        if not self.input_tokens and not self.output_tokens:
            meta = None
        else:
            meta = ApiMeta(
                billed_units=ApiMetaBilledUnits(
                    input_tokens=self.input_tokens, output_tokens=self.output_tokens
                )
            )
        completion = NonStreamedChatResponse(
            generation_id=self.id,
            text=self.message_param.message,
            meta=meta,
            finish_reason=self.finish_reasons[0] if self.finish_reasons else None,
        )

        return CohereCallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



"""A module for interacting with Mirascope RAG."""

from .chunkers import BaseChunker, TextChunker
from .document import Document
from .embedders import BaseEmbedder
from .embedding_params import BaseEmbeddingParams
from .embedding_response import BaseEmbeddingResponse
from .query_results import BaseQueryResults
from .vectorstore_params import BaseVectorStoreParams
from .vectorstores import BaseVectorStore

__all__ = [
    "BaseChunker",
    "TextChunker",
    "BaseEmbedder",
    "BaseEmbeddingParams",
    "BaseEmbeddingResponse",
    "BaseQueryResults",
    "BaseVectorStoreParams",
    "BaseVectorStore",
    "Document",
]



"""This module provides a function to parse content parts from a prompt template."""

import re
import urllib.request
from typing import Any, Literal, cast

from typing_extensions import TypedDict

from ..message_param import (
    AudioPart,
    BaseMessageParam,
    ImagePart,
    TextPart,
)
from ._format_template import format_template
from ._get_audio_type import get_audio_type
from ._get_image_type import get_image_type

_PartType = Literal["text", "image", "images", "audio", "audios"]


class _Part(TypedDict):
    template: str
    type: _PartType
    options: dict[str, str] | None


def _parse_parts(template: str) -> list[_Part]:
    # \{ and \} match the literal curly braces.
    #
    # ([^:{}]+) captures one or more characters that are not { or } or :.
    # This captures the content before the colon.
    #
    # : matches the literal colon separating the type from the content.
    #
    # (image|images|audio|audios) captures the supported special type after the colon.
    #
    # (?:\(([^)]*)\))? captures the optional additional options in parentheses.
    pattern = r"\{([^:{}]+):(image|images|audio|audios)(?:\(([^)]*)\))?\}"
    split = re.split(pattern, template)
    parts: list[_Part] = []
    for i in range(0, len(split), 4):
        if split[i]:
            parts.append(_Part(template=split[i], type="text", options=None))
        if i + 3 < len(split):
            special_content = split[i + 1]
            special_type = cast(_PartType, split[i + 2])
            special_options = split[i + 3]
            if special_options is not None:
                options: dict[str, str] = {}
                for option in special_options.split(","):
                    key, value = option.split("=")
                    options[key] = value
                special_options = options
            parts.append(
                _Part(
                    template=special_content, type=special_type, options=special_options
                )
            )
    return parts


def _load_media(source: str | bytes) -> bytes:
    try:
        # Some typing weirdness here where checking `isinstance(source, bytes)` results
        # in a type hint of `str | bytearray | memoryview` for source in the else.
        if isinstance(source, (bytes, bytearray, memoryview)):
            data = source
        elif source.startswith(("http://", "https://")):
            with urllib.request.urlopen(source) as response:
                data = response.read()
        else:
            with open(source, "rb") as f:
                data = f.read()
        return data
    except Exception as e:  # pragma: no cover
        raise ValueError(
            f"Failed to load or encode data from {source}: {str(e)}"
        )  # pragma: no cover


def _construct_image_part(
    source: str | bytes, options: dict[str, str] | None
) -> ImagePart:
    image = _load_media(source)
    detail = None
    if options:
        detail = options.get("detail", None)
    return ImagePart(
        type="image",
        media_type=f"image/{get_image_type(image)}",
        image=image,
        detail=detail,
    )


def _construct_audio_part(source: str | bytes) -> AudioPart:
    # Note: audio does not currently support additional options, at least for now.
    audio = _load_media(source)
    return AudioPart(
        type="audio", media_type=f"audio/{get_audio_type(audio)}", audio=audio
    )


def _construct_parts(
    part: _Part, attrs: dict[str, Any]
) -> list[TextPart] | list[ImagePart] | list[AudioPart]:
    if part["type"] == "image":
        source = attrs[part["template"]]
        return [_construct_image_part(source, part["options"])] if source else []
    elif part["type"] == "images":
        sources = attrs[part["template"]]
        if not isinstance(sources, list):
            raise ValueError(
                f"When using 'images' template, '{part['template']}' must be a list."
            )
        return (
            [_construct_image_part(source, part["options"]) for source in sources]
            if sources
            else []
        )
    elif part["type"] == "audio":
        source = attrs[part["template"]]
        return [_construct_audio_part(source)] if source else []
    elif part["type"] == "audios":
        sources = attrs[part["template"]]
        if not isinstance(sources, list):
            raise ValueError(
                f"When using 'audios' template, '{part['template']}' must be a list."
            )
        return [_construct_audio_part(source) for source in sources] if sources else []
    else:  # text type
        template = part["template"].strip()
        if not template:
            return []
        return [TextPart(type="text", text=format_template(template, attrs))]


def parse_content_template(
    role: str, template: str, attrs: dict[str, Any]
) -> BaseMessageParam | None:
    """Returns the content template parsed and formatted as a message parameter."""
    if not template:
        return None

    parts = [
        item
        for part in _parse_parts(template)
        for item in _construct_parts(part, attrs)
    ]

    if not parts:
        return None

    if len(parts) == 1 and parts[0].type == "text":
        return BaseMessageParam(role=role, content=parts[0].text)
    return BaseMessageParam(role=role, content=parts)



"""Utility for converting `BaseMessageParam` to `ChatCompletionMessageParam`"""

from cohere.types import ChatMessage

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ChatMessage],
) -> list[ChatMessage]:
    converted_message_params = []
    for message_param in message_params:
        if isinstance(message_param, ChatMessage):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(
                ChatMessage(
                    role=message_param.role.upper(),  # type: ignore
                    message=content,
                )
            )
        else:
            if len(content) != 1 or content[0].type != "text":
                raise ValueError("Cohere does not currently support multimodalities.")
            converted_message_params.append(
                ChatMessage(
                    role=message_param.role.upper(),  # type: ignore
                    message=content[0].text,
                )
            )
    return converted_message_params



"""This module contains the `format_template` function."""

import inspect
from typing import Any

from ._get_template_values import get_template_values
from ._get_template_variables import get_template_variables


def format_template(template: str, attrs: dict[str, Any]) -> str:
    """Formats the given prompt `template`

    Args:
        template: The template to format.
        attrs: The attributes to use for formatting.

    Returns:
        The formatted template.

    """
    dedented_template = inspect.cleandoc(template).strip()
    template_vars = get_template_variables(dedented_template, True)

    values = get_template_values(template_vars, attrs)

    # Remove any special format specs that are actually invalid normally
    dedented_template = dedented_template.replace(":lists", "").replace(":list", "")

    return dedented_template.format(**values).strip()



"""This module contains the `CohereCallResponse` class."""

from cohere.types import (
    ApiMetaBilledUnits,
    ChatMessage,
    NonStreamedChatResponse,
    ToolResult,
)
from pydantic import SkipValidation, computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import CohereCallParams
from .dynamic_config import CohereDynamicConfig
from .tool import CohereTool


class CohereCallResponse(
    BaseCallResponse[
        SkipValidation[NonStreamedChatResponse],
        CohereTool,
        CohereDynamicConfig,
        SkipValidation[ChatMessage],
        CohereCallParams,
        SkipValidation[ChatMessage],
    ]
):
    '''A convenience wrapper around the Cohere `ChatCompletion` response.

    When calling the Cohere API using a function decorated with `cohere_call`, the
    response will be an `CohereCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.cohere import cohere_call

    @cohere_call(model="gpt-4o")
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    response = recommend_book("fantasy")  # response is an `CohereCallResponse` instance
    print(response.content)
    #> Sure! I would recommend...
    ```
    '''

    _provider = "cohere"

    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        return self.response.text

    @property
    def finish_reasons(self) -> list[str] | None:
        """Returns the finish reasons of the response."""
        return [str(self.response.finish_reason)]

    @property
    def model(self) -> str:
        """Returns the name of the response model.

        Cohere does not return model, so we return the model provided by the user.
        """
        return self._model

    @property
    def id(self) -> str | None:
        """Returns the id of the response."""
        return self.response.generation_id

    @property
    def usage(self) -> ApiMetaBilledUnits | None:
        """Returns the usage of the response."""
        if self.response.meta:
            return self.response.meta.billed_units
        return None

    @property
    def input_tokens(self) -> float | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.input_tokens
        return None

    @property
    def output_tokens(self) -> float | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.output_tokens
        return None

    @property
    def cost(self) -> float | None:
        """Returns the cost of the response."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> ChatMessage:
        """Returns the assistant's response as a message parameter."""
        return ChatMessage(
            message=self.response.text,
            tool_calls=self.response.tool_calls,
            role="assistant",  # type: ignore
        )

    @computed_field
    @property
    def tools(self) -> list[CohereTool] | None:
        """Returns the tools for the 0th choice message.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        if not self.tool_types or not self.response.tool_calls:
            return None
        extracted_tools: list[CohereTool] = []
        for tool_call in self.response.tool_calls:
            for tool_type in self.tool_types:
                if tool_call.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break
        return extracted_tools

    @computed_field
    @property
    def tool(self) -> CohereTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tools = self.tools
        if tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls,
        tools_and_outputs: list[tuple[CohereTool, str]],
    ) -> list[ToolResult]:
        """Returns the tool message parameters for tool call results."""
        return [
            ToolResult(
                call=tool.tool_call,
                outputs=[{"output": output}],
            )
            for tool, output in tools_and_outputs
        ]



"""The Mirascope Cohere Module."""

from ._call import cohere_call
from ._call import cohere_call as call
from .call_params import CohereCallParams
from .call_response import CohereCallResponse
from .call_response_chunk import CohereCallResponseChunk
from .dynamic_config import CohereDynamicConfig
from .stream import CohereStream
from .tool import CohereTool

__all__ = [
    "call",
    "CohereDynamicConfig",
    "CohereCallParams",
    "CohereCallResponse",
    "CohereCallResponseChunk",
    "CohereStream",
    "CohereTool",
    "cohere_call",
]



"""Calculate the cost of a completion using the OpenAI API."""


def calculate_cost(
    input_tokens: int | float | None,
    output_tokens: int | float | None,
    model="gpt-3.5-turbo-16k",
) -> float | None:
    """Calculate the cost of a completion using the OpenAI API.

    https://openai.com/pricing

    Model                   Input               Output
    gpt-4o-mini             $0.15 / 1M tokens   $0.60  / 1M tokens
    gpt-4o-mini-2024-07-18  $0.15 / 1M tokens   $0.60  / 1M tokens
    gpt-4o                  $5.00 / 1M tokens   $15.00 / 1M tokens
    gpt-4o-2024-05-13       $5.00 / 1M tokens   $15.00 / 1M tokens
    gpt-4-turbo             $10.00 / 1M tokens  $30.00 / 1M tokens
    gpt-4-turbo-2024-04-09  $10.00 / 1M tokens  $30.00 / 1M tokens
    gpt-3.5-turbo-0125	    $0.50 / 1M tokens	$1.50 / 1M tokens
    gpt-3.5-turbo-1106	    $1.00 / 1M tokens	$2.00 / 1M tokens
    gpt-4-1106-preview	    $10.00 / 1M tokens 	$30.00 / 1M tokens
    gpt-4	                $30.00 / 1M tokens	$60.00 / 1M tokens
    text-embedding-3-small	$0.02 / 1M tokens
    text-embedding-3-large	$0.13 / 1M tokens
    text-embedding-ada-0002	$0.10 / 1M tokens
    """
    pricing = {
        "gpt-4o-mini": {
            "prompt": 0.000_000_15,
            "completion": 0.000_000_6,
        },
        "gpt-4o-mini-2024-07-18": {
            "prompt": 0.000_000_15,
            "completion": 0.000_000_6,
        },
        "gpt-4o": {
            "prompt": 0.000_005,
            "completion": 0.000_015,
        },
        "gpt-4o-2024-05-13": {
            "prompt": 0.000_005,
            "completion": 0.000_015,
        },
        "gpt-4-turbo": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-4-turbo-2024-04-09": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-3.5-turbo-0125": {
            "prompt": 0.000_000_5,
            "completion": 0.000_001_5,
        },
        "gpt-3.5-turbo-1106": {
            "prompt": 0.000_001,
            "completion": 0.000_002,
        },
        "gpt-4-1106-preview": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-4": {
            "prompt": 0.000_003,
            "completion": 0.000_006,
        },
        "gpt-3.5-turbo-4k": {
            "prompt": 0.000_015,
            "completion": 0.000_02,
        },
        "gpt-3.5-turbo-16k": {
            "prompt": 0.000_003,
            "completion": 0.000_004,
        },
        "gpt-4-8k": {
            "prompt": 0.000_003,
            "completion": 0.000_006,
        },
        "gpt-4-32k": {
            "prompt": 0.000_006,
            "completion": 0.000_012,
        },
        "text-embedding-3-small": {
            "prompt": 0.000_000_02,
            "completion": 0.000_000_02,
        },
        "text-embedding-ada-002": {
            "prompt": 0.000_000_1,
            "completion": 0.000_000_1,
        },
        "text-embedding-3-large": {
            "prompt": 0.000_000_13,
            "completion": 0.000_000_13,
        },
    }
    if input_tokens is None or output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_cost = input_tokens * model_pricing["prompt"]
    completion_cost = output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    return total_cost



"""The Mirascope Anthropic Module."""

from ._call import anthropic_call
from ._call import anthropic_call as call
from .call_params import AnthropicCallParams
from .call_response import AnthropicCallResponse
from .call_response_chunk import AnthropicCallResponseChunk
from .dynamic_config import AnthropicDynamicConfig
from .stream import AnthropicStream
from .tool import AnthropicTool

__all__ = [
    "call",
    "AnthropicDynamicConfig",
    "AnthropicCallParams",
    "AnthropicCallResponse",
    "AnthropicCallResponseChunk",
    "AnthropicStream",
    "AnthropicTool",
    "anthropic_call",
]



"""The parameters to use when calling the Gemini API."""

from __future__ import annotations

from google.generativeai.types import GenerationConfig, RequestOptions
from google.generativeai.types.content_types import ToolConfigType
from google.generativeai.types.safety_types import SafetySettingOptions
from typing_extensions import NotRequired

from ..base import BaseCallParams


class GeminiCallParams(BaseCallParams):
    """The parameters to use when calling the Gemini API."""

    generation_config: NotRequired[GenerationConfig]
    safety_settings: NotRequired[SafetySettingOptions]
    request_options: NotRequired[RequestOptions]
    tool_config: NotRequired[ToolConfigType]



"""This module contains the base classes for streaming responses from LLMs."""

import datetime
import inspect
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Generator
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Generic,
    ParamSpec,
    TypeVar,
    overload,
)

from ._utils import (
    HandleStream,
    HandleStreamAsync,
    SetupCall,
    get_fn_args,
    get_metadata,
    get_possible_user_message_param,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_UserMessageParamT = TypeVar("_UserMessageParamT")
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_ToolMessageParamT = TypeVar("_ToolMessageParamT")
_MessageParamT = TypeVar("_MessageParamT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_FinishReason = TypeVar("_FinishReason")


class BaseStream(
    Generic[
        _BaseCallResponseT,
        _BaseCallResponseChunkT,
        _UserMessageParamT,
        _AssistantMessageParamT,
        _ToolMessageParamT,
        _MessageParamT,
        _BaseToolT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _FinishReason,
    ],
    ABC,
):
    """A base class for streaming responses from LLMs."""

    stream: (
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ]
    )
    metadata: Metadata
    tool_types: list[type[_BaseToolT]] | None
    call_response_type: type[_BaseCallResponseT]
    model: str
    prompt_template: str | None = None
    fn_args: dict[str, Any] | None = None
    dynamic_config: _BaseDynamicConfigT
    messages: list[_MessageParamT]
    call_params: _BaseCallParamsT
    call_kwargs: dict[str, Any]
    user_message_param: _UserMessageParamT | None = None
    message_param: _AssistantMessageParamT
    input_tokens: int | float | None = None
    output_tokens: int | float | None = None
    id: str | None = None
    finish_reasons: list[_FinishReason] | None = None
    start_time: float = 0
    end_time: float = 0

    _provider: ClassVar[str] = "NO PROVIDER"

    def __init__(
        self,
        *,
        stream: Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ],
        metadata: Metadata,
        tool_types: list[type[_BaseToolT]] | None,
        call_response_type: type[_BaseCallResponseT],
        model: str,
        prompt_template: str,
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        messages: list[_MessageParamT],
        call_params: _BaseCallParamsT,
        call_kwargs: dict[str, Any],
    ):
        """Initializes an instance of `BaseStream`."""
        self.stream = stream
        self.metadata = metadata
        self.tool_types = tool_types
        self.call_response_type = call_response_type
        self.model = model
        self.prompt_template = prompt_template
        self.fn_args = fn_args
        self.dynamic_config = dynamic_config
        self.messages = messages
        self.call_params = call_params
        self.call_kwargs = call_kwargs
        self.user_message_param = get_possible_user_message_param(messages)  # type: ignore

    def __iter__(
        self,
    ) -> Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]:
        """Iterator over the stream and stores useful information."""
        assert isinstance(
            self.stream, Generator
        ), "Stream must be a generator for __iter__"
        content, tool_calls = "", []
        self.start_time = datetime.datetime.now().timestamp() * 1000
        for chunk, tool in self.stream:
            content += chunk.content
            self._update_properties(chunk)
            if tool:
                tool_calls.append(tool.tool_call)  # type: ignore
            yield chunk, tool
        self.end_time = datetime.datetime.now().timestamp() * 1000
        self.message_param = self._construct_message_param(tool_calls or None, content)

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None]:
        """Iterates over the stream and stores useful information."""

        async def generator():
            assert isinstance(
                self.stream, AsyncGenerator
            ), "Stream must be an async generator for __aiter__"
            content, tool_calls = "", []
            async for chunk, tool in self.stream:
                content += chunk.content
                self._update_properties(chunk)
                if tool:
                    tool_calls.append(tool.tool_call)  # type: ignore
                yield chunk, tool
            self.message_param = self._construct_message_param(
                tool_calls or None, content
            )

        return generator()

    def _update_properties(self, chunk: _BaseCallResponseChunkT):
        """Updates the properties of the stream."""
        if chunk.input_tokens is not None:
            self.input_tokens = (
                chunk.input_tokens
                if not self.input_tokens
                else self.input_tokens + chunk.input_tokens
            )
        if chunk.output_tokens is not None:
            self.output_tokens = (
                chunk.output_tokens
                if not self.output_tokens
                else self.output_tokens + chunk.output_tokens
            )
        if chunk.model is not None:
            self.model = chunk.model
        if chunk.id is not None:
            self.id = chunk.id
        if chunk.finish_reasons is not None:
            self.finish_reasons = chunk.finish_reasons

    @property
    @abstractmethod
    def cost(self) -> float | None:
        """Returns the cost of the stream."""
        ...  # pragma: no cover

    @abstractmethod
    def _construct_message_param(
        self, tool_calls: list[Any] | None = None, content: str | None = None
    ) -> _AssistantMessageParamT:
        """Constructs the assistant message."""
        ...  # pragma: no cover

    def tool_message_params(self, tools_and_outputs) -> list[_ToolMessageParamT]:
        """Returns the tool message parameters for tool call results."""
        return self.call_response_type.tool_message_params(tools_and_outputs)

    @abstractmethod
    def construct_call_response(self) -> _BaseCallResponseT:
        """Constructs the call response."""
        ...  # pragma: no cover


_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_P = ParamSpec("_P")


def stream_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TStream: type[BaseStream],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[TStream]]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream | Awaitable[TStream]]:
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            assert SetupCall.fn_is_sync(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            def generator():
                for chunk, tool in handle_stream(
                    create(stream=True, **call_kwargs), tool_types
                ):
                    yield chunk, tool

            return TStream(
                stream=generator(),
                metadata=get_metadata(fn, dynamic_config),
                tool_types=tool_types,  # type: ignore
                call_response_type=TCallResponse,
                model=model,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                call_kwargs=call_kwargs,
            )

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            assert SetupCall.fn_is_async(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            async def generator():
                async for chunk, tool in handle_stream_async(
                    await create(stream=True, **call_kwargs), tool_types
                ):
                    yield chunk, tool

            return TStream(
                stream=generator(),
                metadata=get_metadata(fn, dynamic_config),
                tool_types=tool_types,  # type: ignore
                call_response_type=TCallResponse,
                model=model,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                call_kwargs=call_kwargs,
            )

        return inner_async if is_async else inner

    return decorator



"""The `gemini_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .stream import GeminiStream
from .tool import GeminiTool

gemini_call = call_factory(
    TCallResponse=GeminiCallResponse,
    TCallResponseChunk=GeminiCallResponseChunk,
    TDynamicConfig=GeminiDynamicConfig,
    TStream=GeminiStream,
    TToolType=GeminiTool,
    TCallParams=GeminiCallParams,
    default_call_params=GeminiCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the Gemini API with a typed function.

This decorator is used to wrap a typed function that calls the Gemini API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@gemini_call(model="gemini-1.5-pro")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The Gemini model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the Gemini API call.
    **call_params: The `GeminiCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an Gemini API call.
'''
